from fnmatch import fnmatchcase
from itertools import combinations
from typing import Callable, Sequence, Union

import numpy as np

from ._data_obj import (
    NumericArg, CategorialArg, CellArg, IndexArg,
    NDVar, Case, Dataset,
    ascategorial, asdataobject, assub,
    cellname, dataobj_repr,
)
from ._stats.stats import SEM, DispersionSpec, variability
from functools import cached_property
from ._utils.numpy_utils import FULL_SLICE


class Celltable:
    """Divide y into cells defined by x.

    Parameters
    ----------
    y : data-object
        dependent measurement
    x : categorial
        Model (Factor or Interaction) for dividing y.
    match : categorial
        Factor on which cases are matched (i.e. subject for a repeated
        measures comparisons). If several data points with the same
        case fall into one cell of x, they are combined using
        match_func. If match is not None, Celltable.groups contains the
        {Xcell -> [match values of data points], ...} mapping corres-
        ponding to self.data
    sub : bool array
        Bool array of length N specifying which cases to include
    cat : None | sequence of cells of x
        Only retain data for these cells. Data will be sorted in the order
        of cells occuring in cat.
    data : Dataset
        If a Dataset is specified, input items (y / x / match / sub) can
        be str instead of data-objects, in which case they will be
        retrieved from the Dataset.
    coercion : callable
        Function to convert the y parameter to to the dependent varaible
        (default: asdataobject).
    dtype : numpy.dtype
        If specified, ``y`` will be converted to this type.


    Examples
    --------
    Split a repeated-measure variable y into cells defined by the
    interaction of A and B::

        >>> c = Celltable(y, A % B, match=subject)


    Attributes
    ----------
    y : data-object
        ``y`` after evaluating input parameters.
    x : categorial
        ``x`` after evaluating input parameters.
    match : categorial | None
        ``match`` after evaluating input parameters.
    sub : bool array | None
        ``sub`` after evaluating input parameters.
    cells : list of (str | tuple)
        List of all cells in x.
    data : {cell: data}
        Data (``y[index]``) in each cell.
    data_indexes : {cell: index-array}
        For each cell, a boolean-array specifying the index for that cell in
        ``x``.

    **If ``match`` is specified**:

    within : dict(cell1, cell2 -> bool)
        Dictionary that specifies for each cell pair whether the corresponding
        comparison is a repeated-measures or an independent measures
        comparison (only available when the input argument ``match`` is
        specified.
    all_within : bool
        Whether all comparison are repeated-measures comparisons or not.
    groups : dict(cell -> group)
        A slice of the match argument describing the group members for each
        cell.

    """
    def __init__(
            self,
            y: NumericArg,
            x: CategorialArg = None,
            match: CategorialArg = None,
            sub: IndexArg = None,
            cat: Sequence[CellArg] = None,
            data: Dataset = None,
            coercion: Callable = asdataobject,
            dtype: np.dtype = None,
    ):
        self.sub = sub
        sub, n_cases = assub(sub, data, return_n=True)

        if x is None:
            if cat is not None:
                raise TypeError(f"{cat=}: cat is only a valid argument if x is provided")
            y, n_cases = coercion(y, sub, data, n_cases, return_n=True)
        else:
            x, n_cases = ascategorial(x, sub, data, n_cases, return_n=True)
            if cat is not None:
                cat = tuple(cat)
                # reconstruct cat if some cells are provided as None
                is_none = [c is None for c in cat]
                if any(is_none):
                    if len(cat) == len(x.cells):
                        if all(is_none):
                            cat = x.cells
                        else:
                            cells = [c for c in x.cells if c not in cat]
                            cat = tuple([cells.pop(0) if c is None else c for c in cat])
                    else:
                        raise ValueError(f"{cat=}: categories can only be specified as None if all cells in x are used, but there are more than {len(cat)} cells: {x.cells}")

                # make sure all categories are in data
                if missing := [repr(c) for c in cat if c not in x.cells]:
                    raise ValueError(f"{cat=} contains categories that are not in the data: {', '.join(missing)}")

                # apply cat
                sort_idx = x.sort_index(order=cat)
                x = x[sort_idx]
                if sub is None:
                    sub = sort_idx
                else:
                    if sub.dtype.kind == 'b':
                        sub = np.flatnonzero(sub)
                    sub = sub[sort_idx]
            y = coercion(y, sub, data, n_cases)

        sort_idx = aggregate = None
        if match is not None:
            match = ascategorial(match, sub, data, n_cases)
            cell_model = match if x is None else x % match
            if len(cell_model) > len(cell_model.cells):
                aggregate = cell_model

            if aggregate is not None:
                # need to aggregate
                y = y.aggregate(aggregate)
                match = match.aggregate(aggregate)
                if x is not None:
                    x = x.aggregate(aggregate)
                    if cat is not None:
                        sort_idx = x.sort_index(order=cat)
            else:
                sort_idx = cell_model.sort_index()
                if x is not None and cat is not None:
                    X_ = x[sort_idx]
                    sort_X_idx = X_.sort_index(order=cat)
                    sort_idx = sort_idx[sort_X_idx]

            if sort_idx is not None:
                if sort_idx[0] == 0 and np.all(np.diff(sort_idx) == 1):
                    sort_idx = None
                else:
                    y = y[sort_idx]
                    match = match[sort_idx]
                    if x is not None:
                        x = x[sort_idx]

        if dtype is not None and y.x.dtype != dtype:
            y = y.astype(dtype)

        # save args
        self.y = y
        self.x = x
        self.cat = cat
        self.match = match
        self.coercion = coercion.__name__
        self.n_cases = len(y)
        # allow aligning additional variables
        self._n_cases = n_cases
        self._sub = sub
        self._aggregate = aggregate
        self._sort_idx = sort_idx

        # extract cell data
        self.data = {}
        self.data_indexes = {}
        if x is None:
            self.data[None] = y
            self.data_indexes[None] = FULL_SLICE
            self.cells = (None,)
            self.n_cells = 1
            self.all_within = match is not None
            return
        self.cells = cat if cat is not None else x.cells
        self.n_cells = len(self.cells)
        self.groups = {}
        for cell in x.cells:
            idx = x.index_opt(cell)
            self.data_indexes[cell] = idx
            self.data[cell] = y[idx]
            if match is not None:
                self.groups[cell] = match[idx]

        # determine which comparisons are within subject comparisons
        if match is not None:
            self.within = {}
            for cell1, cell2 in combinations(self.cells, 2):
                group1 = self.groups[cell1]
                if len(group1) == 0:
                    continue
                group2 = self.groups[cell2]
                if len(group2) == 0:
                    continue
                elif len(group1) != len(group2):
                    within = False
                else:
                    within = np.all(group1 == group2)
                self.within[cell1, cell2] = within
        else:
            self.within = {(cell1, cell2): False for cell1, cell2 in combinations(x.cells, 2)}
        self.any_within = any(self.within.values())
        self.all_within = all(self.within.values())
        self.within.update({(cell2, cell1): within for (cell1, cell2), within in self.within.items()})

    def __repr__(self):
        args = [dataobj_repr(self.y), dataobj_repr(self.x)]
        if self.match is not None:
            args.append(f"match={dataobj_repr(self.match)}")
        if self.sub is not None:
            args.append(f"sub={dataobj_repr(self.sub)}")
        if self.coercion != 'asdataobject':
            args.append(f"coercion={self.coercion}")
        return f"Celltable({', '.join(args)})"

    def __len__(self):
        return self.n_cells

    def _align_ds(
            self,
            data: Dataset,
            rm: bool = False,
            skip: Sequence = (),
            filter: Callable = None,
    ):
        """Align a Dataset to the celltable"""
        out = Dataset()
        reference_cell = self.cells[0]
        for k, v in data.items():
            if k in skip:
                continue
            elif filter and not filter(v):
                continue
            try:
                values = self._align(v, rm)
            except ValueError:  # aggregating failed
                continue
            reference_v = values.pop(reference_cell)
            if all(np.all(vi == reference_v) for vi in values.values()):
                out[k] = reference_v
        return out

    def _align(
            self,
            y: Sequence,
            rm: bool = False,
            data: Dataset = None,
            coerce: Callable = asdataobject,
    ):
        """Align an additional variable to the celltable

        Parameters
        ----------
        y
            Data-object to align.
        rm
            If the celltable is a repeated-measures celltable, align ``y`` to
            the repeated measures table rather than the long form table.
        """
        y_ = coerce(y, self._sub, data, self._n_cases)
        if self._aggregate is not None:
            y_ = y_.aggregate(self._aggregate)
        if self._sort_idx is not None:
            y_ = y_[self._sort_idx]
        if rm:
            return {cell: y_[index] for cell, index in self.data_indexes.items()}
        else:
            return y_

    def cellname(self, cell, delim=' '):
        """Produce a str label for a cell.

        Parameters
        ----------
        cell : tuple | str
            Cell.
        delim : str
            Interaction cells (represented as tuple of strings) are joined by
            ``delim``.
        """
        return cellname(cell, delim=delim)

    def cellnames(self, delim=' '):
        """Return a list of all cell names as strings.

        See Also
        --------
        .cellname : Produce a str label for a single cell.
        """
        return [cellname(cell, delim) for cell in self.cells]

    def data_for_cell(self, cell):
        """Retrieve data for a cell, allowing advanced cell combinations

        Parameters
        ----------
        cell : str | tuple of str
            Name fo the cell. See notes for special cell names. After a special
            cell is retrieved for the first time it is also add to
            ``self.data``.

        Notes
        -----
        Special cell names can be used to retrieve averages between different
        primary cells. The names should be composed so that a case sensitive
        version of fnmatch will find the source cells. For examples, if all
        cells are ``[('a', '1'), ('a', '2'), ('b', '1'), ('b', '2')]``,
        ``('a', '*')`` will retrieve the average of ``('a', '1')`` and
        ``('a', '2')``.
        """
        if cell in self.data:
            return self.data[cell]

        # find cells matched by `cell`
        if isinstance(cell, str):
            cells = [c for c in self.cells if fnmatchcase(c, cell)]
            name = cell
        else:
            cells = [c for c in self.cells if
                     all(fnmatchcase(c_, cp) for c_, cp in zip(c, cell))]
            name = '|'.join(cell)

        # check that all are repeated measures
        for cell1, cell2 in combinations(cells, 2):
            if not self.within[(cell1, cell2)]:
                err = ("Combinatory cells can only be formed from repeated "
                       "measures cells, %r and %r are not." % (cell1, cell2))
                raise ValueError(err)

        # combine data
        cell0 = cells[0]
        x = np.empty_like(self.data[cell0].x)
        for cell_ in cells:
            x += self.data[cell_].x
        x /= len(cells)
        out = NDVar(x, cell0.dims, {}, name)
        self.data[cell] = out
        return out

    def get_data(self, out=list):
        if out is dict:
            return self.data
        elif out is list:
            return [self.data[cell] for cell in self.cells]

    def _get_func(self, cell, func):
        return self.data[cell].aggregate(func=func).x

    @cached_property
    def _pooled_sem(self):
        return SEM(self.y.x, self.x, self.match)

    def _get_dispersion(
            self,
            cell: CellArg,
            spec: Union[str, DispersionSpec],
            pool: bool,  # pooled variance estimate
    ) -> np.ndarray:
        spec = DispersionSpec.from_string(spec)
        if spec.measure == 'SD':
            out = self.data[cell].x.std(0)
            if spec.multiplier != 1:
                out *= spec.multiplier
            return out
        if pool:
            sem = self._pooled_sem
        else:
            sem = SEM(self.data[cell].x)
        return sem.get(spec)

    def get_statistic(self, func=np.mean):
        """Return a list with ``a * func(data)`` for each data cell.

        Parameters
        ----------
        func : callable | str
            statistics function that is applied to the data. Can be string,
            such as '[x]sem' or '[x]ci', e.g. '2sem'.

        See also
        --------
        .get_statistic_dict : return statistics in a ``{cell: data}`` dict
        """
        if isinstance(func, str):
            var_spec = func

            def func(y):
                return variability(y, None, None, var_spec, False)

        return [func(self.data[cell].x) for cell in self.cells]

    def get_statistic_dict(self, func=np.mean):
        """Return a ``{cell: func(data)}`` dictionary.

        Parameters
        ----------
        func : callable | str
            statistics function that is applied to the data. Can be string,
            such as '[x]sem', '[x]std', or '[x]ci', e.g. '2sem'.

        See Also
        --------
        .get_statistic : statistic in a list
        """
        return dict(zip(self.cells, self.get_statistic(func)))

    def variability(self, error='sem', pool=None):
        """Variability measure

        Parameters
        ----------
        error : str
            Measure of variability. Examples:
            ``sem``: Standard error of the mean (default);
            ``2sem``: 2 standard error of the mean;
            ``ci``: 95% confidence interval;
            ``99%ci``: 99% confidence interval.
        pool : bool
            Pool the errors for the estimate of variability (default is True
            for complete within-subject designs, False otherwise).

        Notes
        -----
        Returns within-subject standard error for complete within-subject
        designs (see Loftus & Masson, 1994).
        """
        match = self.match if self.all_within else None
        if pool is None:
            pool = self.all_within
        x = variability(self.y.x, self.x, match, error, pool)
        if isinstance(self.y, NDVar):
            dims = self.y.dims[1:]
            if not pool:
                dims = (Case,) + dims
            return NDVar(x, dims, error, self.y.info.copy())
        else:
            return x



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Pytest configuration

Skipping slow tests: https://docs.pytest.org/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
"""
import pytest


SKIP_MAYAVI = None


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")
    parser.addoption("--no-gui", action="store_true", default=False, help="skip tests that use GUI functionality")
    parser.addoption("--no-mayavi", action="store_true", default=False, help="skip tests that use mayavi")


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")
    config.addinivalue_line("markers", "framework_build: requires framework build")
    config.addinivalue_line("markers", "mayavi: requires mayavi")


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--runslow"):
        skip_slow = pytest.mark.skip(reason="slow test (set --runslow to run)")
        for item in items:
            if "slow" in item.keywords:
                item.add_marker(skip_slow)
    if config.getoption("--no-gui"):
        skip = pytest.mark.skip(reason="requires framework build")
        for item in items:
            if "framework_build" in item.keywords:
                item.add_marker(skip)
    if config.getoption("--no-mayavi"):
        global SKIP_MAYAVI
        SKIP_MAYAVI = True
        skip = pytest.mark.skip(reason="requires mayavi")
        for item in items:
            if "mayavi" in item.keywords:
                item.add_marker(skip)



# https://packaging.python.org/en/latest/
# https://packaging.python.org/en/latest/guides/modernize-setup-py-project/
from packaging.version import Version
import os
from pathlib import Path
import re
from setuptools import setup, find_packages, Extension

import numpy as np

# Distributing Cython modules
# https://cython.readthedocs.io/en/stable/src/userguide/source_files_and_compilation.html#distributing-cython-modules
try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = False


IS_WINDOWS = os.name == 'nt'
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

# version must be in X.X.X format, e.g., "0.0.3dev"
text = (this_directory / 'eelbrain' / '__init__.py').read_text()
match = re.search(r"__version__ = '([.\w]+)'", text)
if match is None:
    raise ValueError("No valid version string found in:\n\n" + text)
version = match.group(1)
Version(version)  # check that it's a valid version

# Cython extensions
args = {'define_macros': [("NPY_NO_DEPRECATED_API", "NPY_1_11_API_VERSION")]}
ext = '.pyx' if cythonize else '.c'
ext_cpp = '.pyx' if cythonize else '.cpp'
if IS_WINDOWS:
    open_mp = dict(extra_compile_args=['/openmp'])
else:
    open_mp = dict(extra_compile_args=['-fopenmp'], extra_link_args=['-fopenmp'])
extensions = [
    Extension('eelbrain._data_opt', [f'eelbrain/_data_opt{ext}'], **args),
    Extension('eelbrain._trf._boosting_opt', [f'eelbrain/_trf/_boosting_opt{ext}'], **open_mp, **args),
    Extension('eelbrain._stats.connectivity_opt', [f'eelbrain/_stats/connectivity_opt{ext}'], **args),
    Extension('eelbrain._stats.opt', [f'eelbrain/_stats/opt{ext}'], **args),
    Extension('eelbrain._stats.vector', [f'eelbrain/_stats/vector{ext_cpp}'], include_dirs=['dsyevh3C'], **args),
]
if cythonize:
    extensions = cythonize(extensions)

setup(
    name='eelbrain',
    version=version,
    description="MEG/EEG analysis tools",
    url="https://eelbrain.readthedocs.io",
    author="Christian Brodbeck",
    author_email='christianbrodbeck@nyu.edu',
    license='BSD (3-clause)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.8',
    setup_requires=[
        "numpy >= 1.20",
        "cython >= 0.21",
    ],
    install_requires=[
        'appnope',
        'colormath >= 2.1',
        'keyring >= 5',
        'matplotlib >= 3.6',
        'mne >= 0.19',
        'nibabel >= 2.5',
        'numpy >= 1.20',
        'pillow',
        'pymatreader',
        'scipy >= 1.5',
        'tqdm >= 4.40',
        'wxpython > 4.1.0',
    ],
    extras_require={
        'brain': [
            'pysurfer[save_movie]',
        ],
    },
    include_dirs=[np.get_include()],
    packages=find_packages(),
    ext_modules=extensions,
    scripts=['bin/eelbrain'],
)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import numpy as np

from .. import fmtxt
from .. import plot
from .. import test
from .. import testnd
from .._data_obj import Dataset, combine
from .._stats.stats import ttest_t
from .._text import ms
from ..fmtxt import Section, linebreak


def source_results(res, surfer_kwargs={}, title="Results", diff_cmap=None,
                   table_pmax=0.2, plot_pmax=0.05):
    "Only used for TRF-report"
    sec = Section(title)

    # raw difference
    brain = plot.brain.surfer_brain(res.difference, diff_cmap, **surfer_kwargs)
    cbar = brain.plot_colorbar(orientation='vertical', show=False)
    sec.add_figure("Correlation increase.", (brain.image('correlation'), cbar))
    brain.close()
    cbar.close()

    # test of difference
    if res._kind == 'cluster':
        clusters = res.find_clusters(table_pmax)
        pmax_repr = str(table_pmax)[1:]
        ctable = clusters.as_table(midrule=True, count=True, caption="All "
                                   "clusters with p<=%s." % pmax_repr)
        sec.append(ctable)

        clusters = res.find_clusters(plot_pmax, True)
        for cluster in clusters.itercases():
            # only plot relevant hemisphere
            sec.add_figure("Cluster %i: p=%.3f" % (cluster['id'], cluster['p']),
                           source_cluster_im(cluster['cluster'], surfer_kwargs),
                           {'class': 'float'})
        sec.append(linebreak)
    return sec


def source_cluster_im(ndvar, surfer_kwargs, mark_sources=None):
    """Plot ('source',) NDVar, only plot relevant hemi

    Parameters
    ----------
    ndvar : NDVar (source,)
        Source space data.
    surfer_kwargs : dict
        Keyword arguments for PySurfer plot.
    mark_sources : SourceSpace index
        Sources to mark on the brain plot (as SourceSpace index).
    """
    kwargs = surfer_kwargs.copy()
    if not ndvar.sub(source='lh').any():
        kwargs['hemi'] = 'rh'
    elif not ndvar.sub(source='rh').any():
        kwargs['hemi'] = 'lh'
    if ndvar.x.dtype.kind == 'b':
        brain = plot.brain.dspm(ndvar, 0, 1.5, **kwargs)
    elif ndvar.x.dtype.kind == 'i':  # map of cluster ids
        brain = plot.brain.surfer_brain(ndvar, 'jet', **kwargs)
    else:
        brain = plot.brain.cluster(ndvar, **kwargs)

    # mark sources on the brain
    if mark_sources is not None:
        mark_sources = np.atleast_1d(ndvar.source._array_index(mark_sources))
        i_hemi_split = np.searchsorted(mark_sources, ndvar.source.lh_n)
        lh_indexes = mark_sources[:i_hemi_split]
        if lh_indexes:
            lh_vertices = ndvar.source.lh_vertices[lh_indexes]
            brain.add_foci(lh_vertices, True, hemi='lh', color="gold")
        rh_indexes = mark_sources[i_hemi_split:]
        if rh_indexes:
            rh_vertices = ndvar.source.rh_vertices[rh_indexes - ndvar.source.lh_n]
            brain.add_foci(rh_vertices, True, hemi='rh', color="gold")

    out = brain.image(ndvar.name)
    brain.close()
    return out


def source_time_results(
        res: testnd.NDTest,
        data: Dataset,
        colors: dict = None,
        include: float = 0.1,
        surfer_kwargs: dict = {},
        title: str = None,
        parc: bool = True,
        y: str = None,
):
    report = Section(title)
    y = data[y or res.y]
    if parc is True:
        parc = res._first_cdist.parc
    model = res._plot_model()
    if parc and res._kind == 'cluster':
        source_bin_table(report, res, surfer_kwargs)

        # add subsections for individual labels
        title = "{tstart}-{tstop} p={p}{mark} {effect}"
        for label in y.source.parc.cells:
            section = report.add_section(label.capitalize())
            clusters = res.find_clusters(source=label)
            source_time_clusters(section, clusters, y, data, model, include, title, colors, res, surfer_kwargs)
    elif not parc and res._kind == 'cluster':
        source_bin_table(report, res, surfer_kwargs)
        clusters = res.find_clusters()
        clusters.sort('tstart')
        title = "{tstart}-{tstop} {location} p={p}{mark} {effect}"
        source_time_clusters(report, clusters, y, data, model, include, title, colors, res, surfer_kwargs)
    elif not parc and res._kind in ('raw', 'tfce'):
        section = report.add_section("P<=.05")
        source_bin_table(section, res, surfer_kwargs, 0.05)
        clusters = res.find_clusters(0.05, maps=True)
        clusters.sort('tstart')
        title = "{tstart}-{tstop} {location} p={p}{mark} {effect}"
        for cluster in clusters.itercases():
            source_time_cluster(section, cluster, y, model, data, title, colors, res.match, surfer_kwargs)

        # trend section
        section = report.add_section("Trend: p<=.1")
        source_bin_table(section, res, surfer_kwargs, 0.1)

        # not quite there section
        section = report.add_section("Anything: P<=.2")
        source_bin_table(section, res, surfer_kwargs, 0.2)
    elif parc and res._kind in ('raw', 'tfce'):
        title = "{tstart}-{tstop} p={p}{mark} {effect}"
        for label in y.source.parc.cells:
            section = report.add_section(label.capitalize())
            # TODO:  **sub is not implemented in find_clusters()
            clusters_sig = res.find_clusters(0.05, True, source=label)
            clusters_trend = res.find_clusters(0.1, True, source=label)
            clusters_trend = clusters_trend.sub("p>0.05")
            clusters_all = res.find_clusters(0.2, True, source=label)
            clusters_all = clusters_all.sub("p>0.1")
            clusters = combine((clusters_sig, clusters_trend, clusters_all))
            clusters.sort('tstart')
            source_time_clusters(section, clusters, y, data, model, include,
                                 title, colors, res, surfer_kwargs)
    else:
        raise RuntimeError
    return report


def source_bin_table(section, res, surfer_kwargs, pmin=1):
    caption = "All clusters in time bins. Each plot shows all sources that are part of a cluster at any time during the relevant time bin. Only the general minimum duration and source number criterion are applied."
    for effect, cdist in res._iter_cdists():
        ndvar = cdist.masked_parameter_map(pmin)
        if not ndvar.any():
            if effect:
                text = f'{effect}: nothing\n'
            else:
                text = 'Nothing\n'
            section.add_paragraph(text)
            continue
        elif effect:
            caption_ = f"{effect}: {caption}"
        else:
            caption_ = caption
        im = plot.brain.bin_table(ndvar, **surfer_kwargs)
        section.add_image_figure(im, caption_)


def source_time_lm(lm, pmin, surfer_kwargs):
    if pmin == 0.1:
        ps = (0.1, 0.01, 0.05)
    elif pmin == 0.05:
        ps = (0.05, 0.001, 0.01)
    elif pmin == 0.01:
        ps = (0.01, 0.0001, 0.001)
    elif pmin == 0.001:
        ps = (0.001, 0.00001, 0.0001)
    else:
        raise ValueError("pmin=%s" % pmin)
    out = Section("SPMs")
    ts = [ttest_t(p, lm.df) for p in ps]
    for term in lm.column_names:
        im = plot.brain.dspm_bin_table(lm.t(term), *ts, summary='extrema',
                                       **surfer_kwargs)
        out.add_section(term, im)
    return out


def source_time_clusters(section, clusters, y, ds, model, include, title,
                         colors, res, surfer_kwargs):
    """Plot cluster with source and time dimensions

    Parameters
    ----------
    ...
    legend : None | fmtxt.Image
        Legend (if shared with other figures).

    Returns
    -------
    legend : fmtxt.Image
        Legend to share with other figures.
    """
    # compute clusters
    if clusters.n_cases == 0:
        section.append("No clusters found.")
        return

    section.append(
        clusters.as_table(midrule=True, count=True, caption="All clusters."))

    # plot individual clusters
    clusters = clusters.sub("p < %s" % include)
    # in non-threshold based tests, clusters don't have unique IDs
    add_cluster_im = 'cluster' not in clusters
    is_multi_effect_result = 'effect' in clusters
    for cluster in clusters.itercases():
        if add_cluster_im:
            if is_multi_effect_result:
                cluster['cluster'] = res.cluster(cluster['id'], cluster['effect'])
            else:
                cluster['cluster'] = res.cluster(cluster['id'])
        source_time_cluster(section, cluster, y, model, ds, title, colors,
                            res.match, surfer_kwargs)


def source_time_cluster(section, cluster, y, model, ds, title, colors, match,
                        surfer_kwargs):
    # cluster properties
    tstart_ms = ms(cluster['tstart'])
    tstop_ms = ms(cluster['tstop'])
    effect = cluster.get('effect', '')

    # section/title
    if title is not None:
        title_ = title.format(tstart=tstart_ms, tstop=tstop_ms,
                              p='%.3f' % cluster['p'], effect=effect,
                              location=cluster.get('location', ''),
                              mark=cluster['sig']).strip()
        while '  ' in title_:
            title_ = title_.replace('  ', ' ')
        section = section.add_section(title_)

    # description
    txt = section.add_paragraph("Id %i" % cluster['id'])
    if 'v' in cluster:
        txt.append(", v=%s" % cluster['v'])
    if 'p_parc' in cluster:
        txt.append(", corrected across all ROIs: ")
        txt.append(fmtxt.eq('p', cluster['p_parc'], 'mcc', '%s', drop0=True))
    txt.append('.')

    # add cluster image to report
    brain = plot.brain.cluster(cluster['cluster'].sum('time'), **surfer_kwargs)
    cbar = brain.plot_colorbar(orientation='vertical', show=False)
    caption = "Cluster"
    if effect:
        caption += 'effect of ' + effect
    caption += "%i - %i ms." % (tstart_ms, tstop_ms)
    section.add_figure(caption, (brain.image('cluster_spatial'), cbar))
    brain.close()
    cbar.close()
    # add cluster time course
    if effect:
        reduced_model = '%'.join(effect.split(' x '))
        if len(reduced_model) < len(model):
            colors_ = plot.colors_for_categorial(ds.eval(reduced_model))
            cluster_timecourse(section, cluster, y, 'source', reduced_model, ds,
                               colors_, match)
    cluster_timecourse(section, cluster, y, 'source', model, ds, colors, match)


def cluster_timecourse(section, cluster, y, dim, model, ds, colors, match):
    c_extent = cluster['cluster']
    cid = cluster['id']

    # cluster time course
    idx = c_extent.any('time')
    tc = y[idx].mean(dim)
    p = plot.UTSStat(tc, model, match=match, data=ds, legend=False, h=4,
                     colors=colors, show=False)
    # mark original cluster
    for ax in p.axes:
        ax.axvspan(cluster['tstart'], cluster['tstop'], color='r',
                   alpha=0.2, zorder=-2)

    if model:
        # legend
        legend_p = p.plot_legend(show=False)
        legend = legend_p.image("Legend")
        legend_p.close()
    else:
        p.axes[0].axhline(0, color='k')
    image_tc = p.image('cluster_%i_timecourse' % cid)
    p.close()

    # Barplot
    idx = (c_extent != 0)
    v = y.mean(idx)
    p = plot.Barplot(v, model, match, data=ds, corr=None, colors=colors, h=4,
                     show=False)
    image_bar = p.image('cluster_%i_barplot.png' % cid)
    p.close()

    # Boxplot
    p = plot.Boxplot(v, model, match, data=ds, corr=None, colors=colors, h=4,
                     show=False)
    image_box = p.image('cluster_%i_boxplot.png' % cid)
    p.close()

    if model:
        # compose figure
        section.add_figure("Time course in cluster area, and average value in cluster by condition, with pairwise t-tests.", [image_tc, image_bar, image_box, legend])
        # pairwise test table
        res = test.pairwise(v, model, match, data=ds, corr=None)
        section.add_figure("Pairwise t-tests of average value in cluster by condition", res)
    else:
        section.add_figure("Time course in cluster area, and average value in cluster.", [image_tc, image_bar, image_box])



# Author:  Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Document model for formatted text documents

This module defines classes for the abstract representation of a document,
which can be rendered into different formats.
Currently (incomplete) support for str, RTF, TeX, PDF (through TeX) and HTML.

Documents are represented with classes inheriting from two base classes:

.. autosummary::
   :toctree: generated

   FMTextElement
   FMText

Subclasses for specific purposes:

.. autosummary::
   :toctree: generated

   Table
   Image
   Figure
   Section
   Report

Functions for export:

.. autosummary::
   :toctree: generated

   copy_pdf
   copy_tex
   save_html
   save_pdf
   save_rtf
   save_tex

"""
from __future__ import annotations

import base64
import csv
import datetime
from html.parser import HTMLParser
from itertools import repeat, zip_longest
from math import ceil
import os
from pathlib import Path
import re
import shutil
import socket
import sys
from io import BytesIO, StringIO
import tempfile
import time
from types import MappingProxyType
from typing import Any, Union, Iterable, List as ListType, Sequence, Tuple
import webbrowser

import numpy as np
import matplotlib.figure
from matplotlib.image import imsave
from matplotlib.mathtext import math_to_image

from ._types import PathArg
from ._utils.tex import latex2pdf
from ._utils import ui


# types
FMTextLike = Union['FMTextElement', str, float, Sequence]  # should be Tuple or List['FMTextLike'], but that causes https://github.com/tox-dev/sphinx-autodoc-typehints/issues/223

ENV = MappingProxyType({})
preferences = dict(
    html_tables_in_fig=True,
)

# Tags
# ----
# Formatting instructions, actions depend on output format.
# Two ways of dealing with them:
#  - substitute tag with X_TAGS[tag]
#  - format native format string with content (X_SUBS[tag] % content)
#
# str:
#     ``sub % content`` if the tag is in _STR_SUBS, otherwise it is
#     ignored
# TeX:
#     Ignor a tag if it is in _TEX_IGNORE;
#     ``sub % content`` if the tag is in _TEX_SUBS,
#     otherwise ``"%s{%s}" % (tag, content)``
# HTML:
#     Formatted HTML if the tag is in _HTML_TAGS, otherwise it is ignored
# RTF:
#     Formatted RTF if the tag is in _RTF_SUBS, otherwise it is ignored

_STR_SUBS = {r'_': '(%s)'}

_TEX_IGNORE = ('font',)
_TEX_SUBS = {'paragraph': "\n\n%s\n\n",
             'math': "$%s$"}  # LaTeX math but normal HTML

_HTML_TAGS = {r'_': 'sub',
              r'^': 'sup',
              r'\author': 'author',
              r'\emph': 'em',
              r'\textbf': 'b',
              r'\textit': 'i',
              'paragraph': 'p',
              'code': 'code',
              'font': 'font'}

_RTF_SUBS = {r'\emph': r"\i %s\i0"}

_html_alignments = {
    'l': 'left',
    'r': 'right',
    'c': 'center',
}

_html_reserved_chars = {'<': '&lt;', '>': '&gt;', '&': '&amp;'}
_html_escape_pattern = re.compile('|'.join(map(re.escape, _html_reserved_chars)))

_tex_escape_chars = {
    '&': r'\&',
    '%': r'\%',
    '$': r'\$',
    '#': r'\#',
    '_': r'\_',
    '{': r'\{',
    '}': r'\}',
    '[': '{[}',
    ']': '{]}',
    '"':  "{''}",
    '~': r'\textasciitilde{}',
    '^': r'\^{}',
    '\\': r'\textbackslash{}',
    '<': r'\textless{}',
    '>': r'\textgreater{}',
}
_tex_escape_pattern = re.compile('|'.join(map(re.escape, _tex_escape_chars)))


def _html_repl(m):
    return _html_reserved_chars[m.group(0)]


def escape_html(text):
    return _html_escape_pattern.sub(_html_repl, text)


def _tex_repl(m):
    return _tex_escape_chars[m.group(0)]


def escape_tex(text):
    return _tex_escape_pattern.sub(_tex_repl, text)


STYLE = """
.float {
    float:left
}
"""

_html_doc_template = """<!DOCTYPE html>
<html>
<head>
    {meta}<title>{title}</title>{style}
</head>

<body>

{body}

</body>
</html>
"""


def get_pdf(tex_obj):
    "Generate PDF from an FMText object (using :mod:`tex`)"
    if isinstance(tex_obj, (Report, Section)):
        doc_class = '{article}'
        standalone = False
    else:
        doc_class = '[border=2pt]{standalone}'
        standalone = True
    txt = tex(tex_obj, {'standalone': standalone})
    document = """
\\documentclass%s
\\usepackage{booktabs}
\\begin{document}
%s
\\end{document}
""" % (doc_class, txt)
    pdf = latex2pdf(document)
    return pdf


def rtf_document(fmtext):
    return "{\\rtf1\\ansi\\deff0\n\n%s\n}" % fmtext.get_rtf()


def display(
        fmtext: FMTextLike,
        title: FMTextLike = None,
):
    """Display the FMText rendered to HTML in a GUI window

    Parameters
    ----------
    fmtext : FMText
        Object to display.
    title : FMText
        Window title.
    """
    from ._wxgui import get_app
    from ._wxgui.text import HTMLFrame

    if title is None:
        if isinstance(fmtext, Section):
            title = fmtext._heading
        elif isinstance(fmtext, Table):
            title = fmtext._title
    if title is None:
        title = 'FMText'
    else:
        title = str(title)

    get_app()
    HTMLFrame(None, title, html(fmtext))


def save_html(
        fmtext: FMTextLike,
        path: PathArg = None,
        embed_images: bool = True,
        meta: dict = None,
):
    """Save an FMText object in HTML format

    Parameters
    ----------
    fmtext : FMText
        Object to save.
    path
        Destination filename. If unspecified, a file dialog will open to ask
        for a destination.
    embed_images
        Embed images in the HTML file (default True). If False, a separate
        folder containing image files is created.
    meta
        Meta-information for document head.
    """
    if path is None:
        msg = "Save as HTML"
        path = ui.ask_saveas(msg, msg, [('HTML (*.html)', '*.html')])
        if not path:
            return
    html_path = Path(path).resolve()
    root = html_path.parent

    if html_path.suffix != '.html':
        html_path = html_path.with_suffix('.html')

    if embed_images:
        resource_dirname = None
    else:
        resource_dirname = html_path.stem
        resource_dir = html_path.with_name(html_path.stem)
        if resource_dir.exists():
            shutil.rmtree(resource_dir)

    buf = make_html_doc(fmtext, root, resource_dirname, meta=meta)
    buf_enc = buf.encode('ascii', 'xmlcharrefreplace')
    html_path.write_bytes(buf_enc)


def save_pdf(
        fmtext: FMTextLike,
        path: PathArg = None,
):
    """Save an FMText object as a PDF (requires LaTeX installation)

    Parameters
    ----------
    fmtext : FMText
        Object to save.
    path
        Destination filename. If unspecified, a file dialog will open to ask
        for a destination.
    """
    pdf = get_pdf(fmtext)
    if path is None:
        msg = "Save as PDF"
        path = ui.ask_saveas(msg, msg, [('PDF (*.pdf)', '*.pdf')])
    if path:
        with open(path, 'wb') as f:
            f.write(pdf)


def save_rtf(
        fmtext: FMTextLike,
        path: PathArg = None,
):
    """Save an FMText object in Rich Text format

    Parameters
    ----------
    fmtext : FMText
        Object to save.
    path
        Destination filename. If unspecified, a file dialog will open to ask
        for a destination.
    """
    text = rtf_document(fmtext)
    if path is None:
        path = ui.ask_saveas("Save RTF", filetypes=[('Rich Text File (*.rtf)', '*.rtf')])
    if path:
        with open(path, 'w') as fid:
            fid.write(text)


def save_tex(
        fmtext: FMTextLike,
        path: PathArg = None,
):
    """Save an FMText object as TeX code

    Parameters
    ----------
    fmtext : FMText
        Object to save.
    path
        Destination filename. If unspecified, a file dialog will open to ask
        for a destination.
    """
    txt = tex(fmtext)
    if path is None:
        path = ui.ask_saveas("Save tex", filetypes=[('tex', 'tex source code')])
    if path:
        with open(path, 'w') as f:
            f.write(txt)


def _save_txt(text, path: PathArg = None):
    if path is None:
        path = ui.ask_saveas("Save Text File", filetypes=[("Plain Text File (*.txt)", "*.txt")])
    if path:
        with open(path, 'w') as fid:
            fid.write(text)


def copy_pdf(fmtext: FMTextLike):
    """Copy an FMText object to the clipboard as PDF

    Parameters
    ----------
    fmtext : FMText
        Object to copy.
    """
    # save pdf to temp file
    pdf = get_pdf(fmtext)
    fd, path = tempfile.mkstemp('.pdf')
    os.write(fd, pdf)
    os.close(fd)

    # copy to clip-board
    ui.copy_file(path)


def copy_tex(fmtext: FMTextLike):
    "Copy an FMText object to the clipboard as tex code"
    txt = tex(fmtext)
    ui.copy_text(txt)


def html(text: FMTextLike, env: dict = ENV):
    """Generate HTML for any object with a string representation

    Parameters
    ----------
    text
        Object to be converted to HTML. If the object has a ``.get_html()``
        method the result of this method is returned, otherwise ``str(text)``.
    env
        Environment for nested HTML.
    """
    if isinstance(text, str):
        return text
    elif hasattr(text, 'get_html'):
        return text.get_html(env)
    else:
        return str(text)


def make_html_doc(
        body: FMTextLike,
        root: str = None,
        resource_dir: str = None,
        title: FMTextLike = None,
        meta: dict = None,
) -> str:
    """Generate HTML document

    Parameters
    ----------
    body
        FMTXT object which should be formatted into an HTML document.
    root
        Path to the directory in which the HTML file is going to be located.
    resource_dir
        Name for the directory containing resources like images, relative to
        root. If None, images are embedded.
    title
        Document title. The default is to try to infer the title from the body
        or use "Untitled".
    meta
        Meta-information for document head.

    Returns
    -------
    html
        HTML document.
    """
    if title is None:
        if hasattr(body, '_site_title') and body._site_title is not None:
            title = html(body._site_title)
        elif hasattr(body, '_heading'):
            title = html(body._heading)
        else:
            title = "Untitled"

    if meta:
        meta = '<meta %s>\n' % ' '.join('%s=%r' % x for x in meta.items())
    else:
        meta = ''

    style = '\n'.join(('', '<style>', STYLE, '</style>'))

    env = {'root': root, 'resource_dir': resource_dir}
    txt_body = html(body, env)
    return _html_doc_template.format(meta=meta, title=title, style=style, body=txt_body)


def tex(text, env=None):
    """Create TeX code for any object with a string representation

    Parameters
    ----------
    text : any
        Object to be converted to HTML. If the object has a ``.get_html()``
        method the result of this method is returned, otherwise ``str(text)``.
    env : dict
        Environment for FMTXT compilation.
    """
    if hasattr(text, 'get_tex'):
        if env is None:
            env = {'math': False}
        elif 'math' not in env:
            env['math'] = False
        return text.get_tex(env)
    else:
        return str(text)


_html_temp = '<{tag}>{body}</{tag}>'
_html_temp_opt = '<{tag} {options}>{body}</{tag}>'


def _html_element(tag: str, body: FMTextLike, env: dict, options: dict = None):
    """Format an HTML element

    Parameters
    ----------
    tag
        The HTML tag.
    body
        The main content between the tags.
    env
        Environment for FMTXT compilation.
    options
        HTML options to be inserted in the start tag.
    """
    if options:
        opt = ' '.join('%s="%s"' % item for item in options.items())
        txt = _html_temp_opt.format(tag=tag, options=opt, body=html(body, env))
    else:
        txt = _html_temp.format(tag=tag, body=html(body, env))
    return txt


def asfmtext(
        content: Any,
        tag: str = None,
        rasterize: bool = None,
        close_figures: bool = None,
):
    """Convert non-FMText objects to FMText

    Parameters
    ----------
    content
        Objects to convert to FMText.
    tag
        Tag to nest ``content`` in.
    rasterize
        Prefer rasterized graphics over vector graphics.
    close_figures
        When encountering figures, close them after rendering. The is mainly
        to prevent the figures to show up separately in notebooks, and is
        enabled by default when in a ``matplotlib inline`` environment.
    """
    if isinstance(content, (FMTextElement, FMTextConstant)):
        if tag:
            return FMTextElement(content, tag)
        else:
            return content
    elif isinstance(content, (list, tuple)):
        return FMText(content, tag)
    elif isinstance(content, matplotlib.figure.Figure):
        from .plot._base import MatplotlibFigure
        content = MatplotlibFigure(content)

    if hasattr(content, '_asfmtext'):
        im = content._asfmtext(rasterize=rasterize, close_figures=close_figures)
        return asfmtext(im, tag)
    else:
        return Text(content, tag)


def asfmtext_or_none(
        content: Any,
        tag: str = None,
        rasterize: bool = None,
):
    "See :func:`asfmtxt`"
    if content is None:
        return None
    else:
        return asfmtext(content, tag, rasterize)


class FMTextConstant:

    def __init__(self, name, html, rtf, tex, text):
        self.name = name
        self.html = html
        self.rtf = rtf
        self.tex = tex
        self.text = text

    def __repr__(self):
        return self.name

    def get_html(self, env):
        return self.html

    def get_rtf(self, env):
        return self.rtf

    def get_tex(self, env):
        return self.tex

    def get_str(self, env):
        return self.text


linebreak = FMTextConstant(
    name='linebreak',
    html='<br style="clear:left">\n',
    rtf='\\line\n',
    tex='\\\\\n',
    text='\n',
)


class FMTextElement:
    """A text element along with formatting specification"""
    def __init__(
            self,
            content: object,
            tag: str = None,
            options: dict = None,
    ):
        """Represent a value along with formatting properties.

        Parameters
        ----------
        content
            Any item with a string representation (str, scalar, ...).
        tag
            Formatting tag.
        options
            Options for HTML tags.
        """
        self.content = content
        self.tag = tag
        self.options = options

    def __repr__(self):
        args = ', '.join(self._repr_items())
        return f"{self.__class__.__name__}({args})"

    def _repr_items(self):
        items = [repr(self.content)]
        if self.tag:
            items.append(repr(self.tag))
            if self.options:
                items.append(repr(self.options))
        return items

    def __str__(self):
        return self.get_str()

    def _repr_html_(self):
        return html(self)

    def __add__(self, other):
        if isinstance(other, str) and other == '':
            # added to prevent matplotlib from thinking Image is a file path
            raise ValueError("Can't add empty string")

        return FMText([self, other])

    def copy_pdf(self):
        "Copy PDF to clipboard"
        copy_pdf(self)

    def copy_tex(self):
        "Copy TeX to clipboard"
        copy_tex(self)

    def _get_core(self, env):
        "Unicode core"
        return str(self.content)

    def get_html(self, env):
        "Complete HTML representation"
        txt = self._get_html_core(env)

        tag = None
        if self.tag:
            if self.tag in _HTML_TAGS:
                tag = _HTML_TAGS[self.tag]
            elif re.match(r'h\d', self.tag):
                tag = self.tag

        if tag:
            return _html_element(tag, txt, env, self.options)
        else:
            return txt

    def _get_html_core(self, env):
        "HTML representation of everything inside the tag"
        return escape_html(self._get_core(env))

    def get_rtf(self, env: dict = ENV) -> str:
        if self.tag in _RTF_SUBS:
            return _RTF_SUBS[self.tag] % self._get_rtf_core(env)
        else:
            return self._get_rtf_core(env)

    def _get_rtf_core(self, env):
        return self._get_core(env)

    def get_str(self, env: dict = ENV) -> str:
        "String representation"
        if self.tag in _STR_SUBS:
            return _STR_SUBS[self.tag] % self._get_str_core(env)
        else:
            return self._get_str_core(env)

    def _get_str_core(self, env):
        return self._get_core(env)

    def get_tex(self, env) -> str:
        "TeX representation"
        if self.tag == 'math':
            if env['math']:
                raise RuntimeError("Nested math tag")
            env['math'] = True
        txt = self._get_tex_core(env)
        if self.tag == 'math':
            env['math'] = False

        if self.tag and self.tag not in _TEX_IGNORE:
            if self.tag in _TEX_SUBS:
                txt = _TEX_SUBS[self.tag] % txt
            else:
                txt = r"%s{%s}" % (self.tag, txt)

        return txt

    def _get_tex_core(self, env):
        out = self._get_core(env)
        if env['math']:
            return out
        else:
            return escape_tex(out)

    def save_html(self, path: PathArg = None, embed_images=True, meta=None):
        """Save in HTML format

        Parameters
        ----------
        path
            Destination filename. If unspecified, a file dialog will open to ask
            for a destination.
        embed_images : bool
            Embed images in the HTML file (default True). If False, a separate
            folder containing image files is created.
        meta : dict
            Meta-information for document head.
        """
        save_html(self, path, embed_images, meta)

    def save_pdf(self, path: PathArg = None):
        """Save in PDF format

        Parameters
        ----------
        path
            Destination filename. If unspecified, a file dialog will open to ask
            for a destination.
        """
        save_pdf(self, path)

    def save_rtf(self, path: PathArg = None):
        """Save in Rich Text format

        Parameters
        ----------
        path
            Destination filename. If unspecified, a file dialog will open to ask
            for a destination.
        """
        save_rtf(self, path)

    def save_tex(self, path: PathArg = None):
        """Save in TeX format

        Parameters
        ----------
        path
            Destination filename. If unspecified, a file dialog will open to ask
            for a destination.
        """
        save_tex(self, path)

    def save_txt(self, path: PathArg = None):
        """Save as plain text file

        Parameters
        ----------
        path
            Destination filename. If unspecified, a file dialog will open to ask
            for a destination.
        """
        _save_txt(self.get_str(), path)


class FMText(FMTextElement):
    r"""List of :class:`FMTextElement` items

    Parameters
    ----------
    content : FMTextLike
        Any item with a string representation (str, FMText, scalar, ...)
        or an object that iterates over such items (e.g. a list of FMText).
    tag : str
        Formatting tag.
    options : dict
        Options for HTML tags.
    rasterize
        Try to rasterize images in ``content``.

    Notes
    -----
    The ``property`` argument is primarily used for TeX commands; for
    example, ``txt = FMText('Brie', r'\textbf')`` will result in the
    following TeX representation: r"\textbf{Brie}". Some properties are
    automatically translated into HTML; thus, the HTML representation of
    ``txt`` would be "<b>Brie</b>". The following are additional proprties
    that don't exist in TeX:

    'paragraph'
        A <p> tag in HTML, and simple line breaks in TeX.
    """
    def __init__(
            self,
            content: FMTextLike = None,
            tag: str = None,
            options: dict = None,
            rasterize: bool = False,
    ):
        if content is None:
            content = []
        elif isinstance(content, (list, tuple)):
            content = [asfmtext(item, rasterize=rasterize) for item in content]
        else:
            content = [asfmtext(content, rasterize=rasterize)]
        FMTextElement.__init__(self, content, tag, options)

    def append(self, content):
        """Append content to the FMText item

        Parameters
        ----------
        content : str | object | iterable
            Any item with a string representation (str, FMText, scalar, ...)
            or an object that iterates over such items (e.g. a list of FMText).
        """
        self.content.append(asfmtext(content))

    def _get_html_core(self, env):
        return ''.join(i.get_html(env) for i in self.content)

    def _get_rtf_core(self, env):
        return ''.join(i.get_rtf(env) for i in self.content)

    def _get_str_core(self, env):
        return ''.join(i.get_str(env) for i in self.content)

    def _get_tex_core(self, env):
        return ''.join(i.get_tex(env) for i in self.content)


class Code(FMTextElement):
    """Code (block) element

    Parameters
    ----------
    code : str
        Multiline string to be displayed as code.
    """
    def __init__(self, content):
        assert isinstance(content, str)
        FMTextElement.__init__(self, content, 'code')

    def get_tex(self, env):
        raise NotImplementedError

    def _get_html_core(self, env):
        return linebreak.html.join(line.rstrip().replace(' ', '&nbsp;') for
                                   line in self.content.splitlines())


class Text(FMTextElement):

    def __init__(self, content, tag=None):
        if not isinstance(content, str):
            content = str(content)
        FMTextElement.__init__(self, content, tag)


class Link(FMTextElement):

    def __init__(self, content: FMTextLike, url: str):
        FMTextElement.__init__(self, content)
        self.url = str(url)

    def get_html(self, env):
        return '<a href="%s">%s</a>' % (self.url, FMTextElement.get_html(self, env))


class Number(FMTextElement):

    def __init__(self, content, tag=None, fmt='%s', drop0=False):
        if not np.isscalar(content):
            if getattr(content, 'ndim', None) == 0:
                content = content.item()
            else:
                raise TypeError("content=%s" % repr(content))
        FMTextElement.__init__(self, content, tag)
        self.fmt = fmt
        self.drop0 = drop0

    def _repr_items(self):
        items = FMTextElement._repr_items(self)
        if self.fmt != '%s':
            items.append('fmt=%r' % self.fmt)
        if self.drop0:
            items.append('drop0=True')
        return items

    def _get_core(self, env):
        if np.isnan(self.content):
            return 'NaN'
        elif isinstance(self.content, (bool, np.bool_)):
            return str(self.content)
        else:
            fmt = env.get('fmt', self.fmt)
            txt = fmt % self.content
            if self.drop0:
                if txt.startswith('0.'):
                    return txt[1:]
                elif txt.startswith('-0.'):
                    return '-' + txt[2:]
            return txt


class P(Number):

    def __init__(self, content: float):
        Number.__init__(self, content, fmt='%.3f', drop0=True)

    def _get_core(self, env):
        if self.content < .001:
            return '< .001'
        else:
            return Number._get_core(self, env)


class Math(FMTextElement):
    r""":class:`FMTextElement` for math expressions

    Parameters
    ----------
    content : str
        LaTeX math expression (excluding '$'). E.g. ``r'\sqrt{a}'``.
    equation : bool
        Whether to display the expression as a separate equation (as
        oppposed to inline).
    """
    def __init__(self, content, equation=False):
        FMTextElement.__init__(self, content)
        self._equation = equation

    def __repr__(self):
        items = [repr(self.content)]
        if self._equation:
            items.append("equation=True")
        return "Math(%s)" % ', '.join(items)

    def get_html(self, env: dict = ENV):
        im = Image("LaTeX Equation", 'svg', self.content)
        math_to_image("$%s$" % self.content, im, format='svg')
        if self._equation:
            return '\n<br>\n%s\n<br>\n' % im.get_html(env)
        else:
            return im.get_html(env)

    def get_tex(self, env):
        if self._equation:
            return "\\begin{equation}$%s$\\end{equation}" % self.content
        else:
            return "$%s$" % self.content


class EquationArray(FMTextElement):
    """:class:`FMTextElement` for equation arrays

    Parameters
    ----------
    eqnarray : tuple of str
        Tuple of lines for the equation array.
    """
    def __init__(self, eqnarray):
        FMTextElement.__init__(self, eqnarray)

    def __repr__(self):
        return "EquationArray(%s)" % repr(self.content)

    def get_html(self, env):
        ims = []
        for line in self.content:
            im = Image("LaTeX EquationArray", 'svg', line)
            math_to_image("$%s$" % line.replace('&', ''), im, format='svg')
            ims.append(im.get_html(env))
        return '\n<br>\n%s\n<br>\n' % '\n<br>\n'.join(ims)

    def get_tex(self, env):
        return "\\begin{eqnarray}%s\\end{eqnarray}" % r'\\'.join(self.content)


class Stars(FMTextElement):
    """:class:`FMTextElement` for decoration of p-values

    Shortcut for adding stars to a table and spaces in place of absent stars,
    so that alignment to the right can be used.
    """
    def __init__(self, n, of=3, tag="^"):
        if isinstance(n, str):
            self.n = len(n.strip())
        else:
            self.n = n
        self.of = of
        if np.isreal(n):
            text = '*' * n + ' ' * (of - n)
        else:
            text = n.ljust(of)
        FMTextElement.__init__(self, text, tag)

    @classmethod
    def from_p(cls, p, of=3, tag='^'):
        n = sum((p <= 0.001, p <= 0.01, p <= 0.05))
        return cls(n, of, tag)

    def _get_tex_core(self, env):
        txt = self._get_core(env)
        spaces = r'\ ' * (self.of - self.n)
        return txt + spaces


class List(FMTextElement):
    """Bulletted list of FMText elements"""
    def __init__(
            self,
            head: FMTextLike = None,
            items: Iterable[FMTextLike] = (),
            ordered: bool = False,
    ):
        """Bulletted list of FMText elements

        Parameters
        ----------
        head : FMTextLike
            First line on higher level (no bullet for highest list, or list
            element for subordinate list).
        items : iterable of FMTextLike
            List items.
        ordered : bool
            Whether to use the "ol" HTML tag (instead of "ul").
        """
        self.ordered = ordered
        self.head = asfmtext_or_none(head)
        self.items = [asfmtext(item) for item in items]

    def _repr_items(self):
        if self.ordered:
            return [repr(self.head), repr(self.items), repr(self.ordered)]
        elif self.items:
            return [repr(self.head), repr(self.items)]
        elif self.head:
            return [repr(self.head)]
        else:
            return []

    def add_item(self, item: FMTextLike):
        "Add an item to the list"
        self.items.append(asfmtext(item))

    def add_sublist(
            self,
            head: FMTextLike = None,
            items: Iterable[FMTextLike] = (),
            ordered: bool = False,
    ):
        """Add an item with a subordinate list

        Parameters
        ----------
        head : FMTextLike
            Text for the parent item
        items : iterable of FMTextLike
            Subordinate list items.
        ordered : None | bool
            Whether to use the "ol" HTML tag (instead of "ul"). The default is
            to inherit the parent list's setting.

        Returns
        -------
        sublist : List
            The subordinate list.
        """
        if ordered is None:
            ordered = self.ordered
        sublist = List(head, items, ordered)
        self.add_item(sublist)
        return sublist

    def get_html(self, env: dict = ENV):
        items = []
        if self.head is not None:
            items.append(self.head.get_html(env))
        tag = 'ol' if self.ordered else 'ul'
        items.append('<%s>' % tag)

        # body
        for item in self.items:
            items.append(_html_element('li', item, env))

        items.append('</%s>' % tag)
        return '\n'.join(items)

    def get_str(self, env: dict = ENV):
        out = []
        if self.head is not None:
            out.append(self.head.get_str(env))

        for item in self.items:
            if isinstance(item, List):
                lines = item.get_str(env).splitlines()
                out.append('- %s' % lines[0])
                out.extend('  %s' % line for line in lines[1:])
            else:
                out.append('- %s' % str(item))
        return '\n'.join(out)


# Table ---

class Cell(FMText):

    def __init__(self, content='', tag=None, width=1, just=None):
        """A cell for a table

        Parameters
        ----------
        text : FMText
            Cell content.
        width : int
            Width in columns for multicolumn cells.
        just : None | 'l' | 'r' | 'c'
            Justification. None: use column standard.
        others :
            FMText parameters.
        """
        FMText.__init__(self, content, tag)
        self.width = width
        if width > 1 and not just:
            self.just = 'l'
        else:
            self.just = just

    @classmethod
    def coerce(cls, content):
        if isinstance(content, cls):
            return content
        return cls(content)

    def _repr_items(self):
        items = FMText._repr_items(self)
        if self.width != 1:
            i = min(2, len(items))
            items.insert(i, 'width=%s' % self.width)
        return items

    def __len__(self):
        return self.width

    def get_html(self, align: str, env: dict = ENV):
        html_repr = FMText.get_html(self, env)
        options = {}
        # width
        if self.width > 1:
            options['colspan'] = self.width
        # alignment
        if self.just:
            align = self.just
        options['style'] = f"text-align:{_html_alignments[align]}"
        return _html_element('td', html_repr, env, options)

    def get_rtf(self, env: dict = ENV):
        return "%s\\intbl\\cell" % FMText.get_rtf(self, env)

    def get_tex(self, env: dict = ENV):
        tex_repr = FMText.get_tex(self, env)
        if self.width > 1 or self.just:
            tex_repr = r"\multicolumn{%s}{%s}{%s}" % (self.width, self.just,
                                                      tex_repr)
        return tex_repr


class Row(list):
    """Row for a Table"""
    def __init__(self, n_columns, items=()):
        self.n_columns = n_columns
        list.__init__(self, (Cell.coerce(item) for item in items))

    @classmethod
    def coerce(cls, obj, n_columns):
        if isinstance(obj, cls):
            if obj.n_columns != n_columns:
                raise ValueError("Required Row with %i columns, got %i" %
                                 (n_columns, obj.n_columns))
            return obj
        return cls(n_columns, obj)

    def __len__(self):
        return sum([len(cell) for cell in self])

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            start, stop, stride = key.indices(self.n_columns)
            for i, v in zip(range(start, stop, stride), value):
                self[i] = v
            return
        elif not isinstance(key, int):
            raise TypeError("Row index %r" % (key,))

        column = key if key >= 0 else key + self.n_columns
        if not 0 <= column < self.n_columns:
            raise IndexError("%i for row with %i columns" % (key, self.n_columns))

        cell = Cell.coerce(value)
        if len(self) <= column:
            while len(self) < column:
                self.append(Cell(''))
            self.append(cell)
            return

        column_i = 0
        for i, cell_i in enumerate(self):
            if column_i == column:
                list.__setitem__(self, i, cell)
                return
            column_i += len(cell_i)
        raise IndexError("Column %i is part of a multi-column cell" % (column,))

    def __repr__(self):
        return "Row(%s)" % list.__repr__(self)

    def __str__(self):
        return ' '.join([str(cell) for cell in self])

    def _col_str_lens(self, env) -> ListType[int]:
        "List of cell-str-lengths; multicolumns are handled poorly"
        lens = []
        for cell in self:
            cell_len = len(cell.get_str(env))
            n_columns = len(cell)
            if n_columns == 1:
                lens.append(cell_len)
            else:
                # TODO: better handling of multicolumn
                col_len = int(ceil(cell_len / n_columns))
                lens.extend(repeat(col_len, n_columns))
        return lens

    def cell(
            self,
            content: FMTextLike = None,
            tag: str = None,
            width: int = 1,
            just: str = None,
    ):
        """Add a cell to the row

        Parameters
        ----------
        content : FMText
            Cell content.
        tag : str
            Formatting tag.
        width : int
            Width in columns for multicolumn cells.
        just : 'l' | 'r' | 'c'
            Justification (default: use column standard).
        """
        cell = Cell(content, tag, width, just)
        if len(cell) + len(self) > self.n_columns:
            raise ValueError(f"width={cell.width}: exceeds table width")
        self.append(cell)

    def cells(self, *cells, **kwargs):
        "Add several simple cells with one command"
        for cell in cells:
            self.cell(cell, **kwargs)

    def get_html(self, c_just: Sequence[str], env: dict = ENV):
        col = 0
        items = ['<tr>']
        for cell in self:
            items.append(cell.get_html(c_just[col], env))
            col += cell.width
        items.append('</tr>')
        return '\n'.join(items)

    def get_rtf(self, env: dict = ENV):
        return '\n'.join([cell.get_rtf(env) for cell in self] + ['\\row'])

    def get_row_strings(
            self,
            c_width: Sequence[int],  # column width
            c_just: Sequence[str],  # global color alignment
            delimiter: str = '   ',
            env: dict = ENV,
    ) -> ListType[str]:
        "String of the row using column spacing ``c_width``"
        # find strings for each column
        col = 0
        col_strings: ListType[ListType[str]] = []
        for cell in self:
            if cell.width == 1:
                target_len = c_width[col]
            else:
                target_len = sum(c_width[col:col + cell.width])
                target_len += len(delimiter) * (cell.width - 1)
            raw_lines = cell.get_str(env).splitlines() or ['']
            # split long lines to fit into column
            lines: ListType[str] = []
            for raw_line in raw_lines:
                if len(raw_line) > target_len:  # TODO: nicer splitting
                    lines += [raw_line[i: i+target_len] for i in range(0, len(raw_line), target_len)]
                else:
                    lines.append(raw_line)
            # justify text
            just = cell.just or c_just[col]
            if just == 'l':
                lines = [line.ljust(target_len) for line in lines]
            elif just == 'r':
                lines = [line.rjust(target_len) for line in lines]
            elif just == 'c':
                rj = target_len // 2
                lines = [line.rjust(rj).ljust(target_len) for line in lines]
            else:
                raise ValueError(f'Justification {just=} encountered')
            col_strings.append(lines)
            col += cell.width

        # fill in spaces for empty cell-rows
        n_rows = max(len(col_i) for col_i in col_strings)
        for lines in col_strings:
            for _ in range(n_rows - len(lines)):
                lines.append(' ' * len(lines[0]))

        return [delimiter.join(items) for items in zip(*col_strings)]

    def get_tex(self, env=None):
        if env is None:
            return tex(self)
        out = ' & '.join(cell.get_tex(env) for cell in self)
        out += r" \\"
        return out


class Table(FMTextElement):
    r"""A table

    Parameters
    ----------
    columns : str
        alignment for each column, e.g. ``'lrr'``
    rules : bool
        Add toprule and bottomrule
    title : text
        Title for the table.
    caption : text
        Caption for the table.
    rows : list of Row
        Table body.

    Examples
    --------
    >>> from eelbrain import fmtxt
    >>> table = fmtxt.Table('lll')
    >>> table.cell()
    >>> table.cell("example 1")
    >>> table.cell("example 2")
    >>> table.midrule()
    >>> table.cell("string")
    >>> table.cell('???')
    >>> table.cell('another string')
    >>> table.cell("Number")
    >>> table.cell(4.5)
    >>> table.cell(2./3, fmt='%.4g')
    >>> print(table)
             example 1   example 2
    -----------------------------------
    string   ???         another string
    Number   4.5         0.6667
    >>> print(table.get_tex())
    \begin{center}
    \begin{tabular}{lll}
    \toprule
     & example 1 & example 2 \\
    \midrule
    string & ??? & another string \\
    Number & 4.5 & 0.6667 \\
    \bottomrule
    \end{tabular}
    \end{center}
    >>> table.save_tex()

    """
    def __init__(self, columns, rules=True, title=None, caption=None, rows=()):
        self.columns = columns
        self.n_columns = len(columns)
        self.rows = list(rows)
        self.rules = rules
        self.title(title)
        self.caption(caption)
        self._active_row = None

    @property
    def shape(self):
        return len(self.rows), self.n_columns

    def __len__(self):
        return len(self.rows)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.rows[item]
        elif isinstance(item, slice):
            rows = self.rows[item]
            return Table(self.columns, rules=self.rules, title=self._title,
                         caption=self._caption, rows=rows)
        raise TypeError(f'Table index {item}')

    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.rows[key] = Row.coerce(value, self.n_columns)
        elif isinstance(key, tuple) and len(key) == 2:
            row, column = key
            if isinstance(row, slice):
                value = tuple(value)
                start, stop, stride = row.indices(len(self.rows))
                for i, v in zip(range(start, stop, stride), value):
                    self[i, column] = v
            elif not isinstance(row, int):
                raise TypeError("Table index %r" % (key,))
            else:
                self.rows[row][column] = value
        else:
            raise IndexError("Table index %r" % (key,))

    # adding texstrs ---
    def cell(
            self,
            content: FMTextLike = None,
            tag: str = None,
            width: int = 1,
            just: str = None,
    ):
        """Add a cell to the table

        Parameters
        ----------
        content : FMText
            Cell content.
        tag : str
            Formatting tag.
        width : int
            Width in columns for multicolumn cells.
        just : 'l' | 'r' | 'c'
            Justification (default: use column standard).
        """
        if self._active_row is None or len(self._active_row) == self.n_columns:
            self._active_row = self.add_row()
        self._active_row.cell(content, tag, width, just)

    def cells(self, *cells, **kwargs):
        "Add several simple cells with one command"
        for cell in cells:
            self.cell(cell, **kwargs)

    def add_row(self, at=None):
        """Add a row without affecting the cursor

        Parameters
        ----------
        at : int
            Index at which to insert row (default is after the current row;
            midrules also count as rows).

        Returns
        -------
        row : Row
            the new row.
        """
        row = Row(self.n_columns)
        if at is None:
            self.rows.append(row)
        else:
            self.rows.insert(at, row)
        return row

    def empty_row(self):
        self.endline()
        self._active_row = self.add_row()

    def endline(self):
        "Finish the active row"
        if self._active_row is not None:
            for _ in range(self.n_columns - len(self._active_row)):
                self._active_row.append(Cell())
            self._active_row = None

    def midrule(self, span=None):
        """Add a midrule

        note that a toprule and a bottomrule are inserted automatically
        in every table.

        Parameters
        ----------
        span : str | sequence of int
            Add a midrule that spans less than the whole width of the table
            (e.g., ``'2-4'`` or ``(2, 4)``).
        """
        self.endline()
        if span is None:
            self.rows.append("\\midrule")
        else:
            if isinstance(span, (list, tuple)):
                span = '%i-%i' % span
            elif isinstance(span, str):
                if not re.match(r'\d+-\d+', span):
                    raise ValueError("span=%r" % span)
            else:
                raise TypeError("span=%r" % (span,))
            self.rows.append(r"\cmidrule{%s}" % span)

    def title(self, content: FMTextLike):
        """Set the table title"""
        self._title = asfmtext_or_none(content)

    def caption(self, content: FMTextLike):
        """Set the table caption"""
        self._caption = asfmtext_or_none(content)

    def append(self, table: 'Table'):
        "Append another table (has to have same column specification)"
        if table.columns != self.columns:
            raise ValueError(f"Trying to append table with table.colums={table.columns} to {self.columns}")
        self.endline()
        self.rows.extend(table.rows)

    def __repr__(self):
        # return self.__str__ so that when a function returns a Table, the
        # result can be inspected without assigning the Table to a variable.
        return self.__str__()

    def get_html(self, env: dict = ENV):
        if self._title is None:
            title = None
        else:
            title = _html_element('h3', self._title, env)

        if self._caption is None:
            caption = None
        else:
            if preferences['html_tables_in_fig']:
                tag = 'figcaption'
            else:
                tag = 'caption'
            caption = _html_element(tag, self._caption, env)

        # table body
        table = []
        if not preferences['html_tables_in_fig']:
            if title:
                table.append(title)
            if caption:
                table.append(caption)
        for row in self.rows:
            if isinstance(row, str):
                if row == "\\midrule":
                    pass
#                     table.append('<tr style="border-bottom:1px solid black">')
            else:
                table.append(row.get_html(self.columns, env))
        body = '\n'.join(table)

        # table frame
        if self.rules:
            table_options = {'border': 1, 'frame': 'hsides', 'rules': 'none'}
        else:
            table_options = {'border': 0}
        table_options['cellpadding'] = 2
        txt = _html_element('table', body, env, table_options)

        # embedd in a figure
        if preferences['html_tables_in_fig']:
            if title or caption:
                txt = '\n'.join([item for item in [title, txt, caption] if item is not None])
            txt = _html_element('figure', txt, env)

        return txt

    def get_rtf(self, env: dict = ENV):
        # header
        rows = [r'\cellx%i000' % i for i in range(len(self.columns))]
        rows.insert(0, '\\trowd')
        rows.append('\\row')
        # body
        for row in self.rows:
            if isinstance(row, str):
                if row == "\\midrule":
                    pass
            else:
                rows.append(row.get_rtf(env))
        return '\n'.join(rows)

    def get_str(
            self,
            env: dict = ENV,
            delim='   ', linesep='\n',
            max_width: int = None,
    ):
        """Convert Table to str

        Parameters
        ----------
        env
            Processing environment.
        delim
            Delimiter between columns.
        linesep
            Line separation string
        max_width
            Maximum table width in characters, or 0 to disable text wrapping
            (make the table as wide as necessary). Default is based on the
            current terminal width (:func:`shutil.get_terminal_size`), unless
            ``delim`` contains the ``tab`` character, in which the obligatory
            default is 0.
        """
        if len(self.rows) == 0:
            return ''
        if max_width is None:
            if '\t' in delim:
                max_width = 0
            else:
                max_width = shutil.get_terminal_size((0, 80))[0]

        # determine ideal column widths
        # TODO: consider multi-column cells separately at the end
        widths = [row._col_str_lens(env) for row in self.rows if not isinstance(row, str)]
        column_widths = [max(cols) for cols in zip_longest(*widths, fillvalue=0)]
        while len(column_widths) < len(self.columns):
            column_widths.append(0)

        # determine table width & actual column widths
        if '\t' in delim:
            if max_width:
                raise ValueError(f'{max_width=} with {delim=}')
            midrule = delim.join(['-' * w for w in column_widths])
        else:
            cumulative_delim_with = len(delim) * (len(self.columns) - 1)
            cumulative_with = cumulative_delim_with + sum(column_widths)
            if max_width and cumulative_with > max_width:
                # determine wrapping
                if max_width < cumulative_delim_with + len(self.columns):
                    raise ValueError(f'{max_width=} results in column width < 1')
                available_width = max_width - cumulative_delim_with
                new_c_width = [0] * len(self.columns)
                i = 0
                while available_width > 0:
                    if new_c_width[i] < column_widths[i]:
                        new_c_width[i] += 1
                        available_width -= 1
                    i = (i + 1) % len(self.columns)
                column_widths = new_c_width
                cumulative_with = max_width
            midrule = '-' * cumulative_with

        # collect lines
        txtlines = []
        for row in self.rows:
            if isinstance(row, str):  # commands
                if row == r'\midrule':
                    txtlines.append(midrule)  # "_"*l_len)
                elif row == r'\bottomrule':
                    txtlines.append(midrule)  # "_"*l_len)
                elif row == r'\toprule':
                    txtlines.append(midrule)  # "_"*l_len)
                elif row.startswith(r'\cmidrule'):
                    txt = row.split('{')[1]
                    txt = txt.split('}')[0]
                    start, end = txt.split('-')
                    start = int(start) - 1
                    end = int(end)
                    line = [' ' * w for w in column_widths[:start]]
                    rule_delim = '-' * len(delim.expandtabs(4))
                    line += [rule_delim.join(['-' * w for w in column_widths[start:end]])]
                    line += [' ' * w for w in column_widths[start:end]]
                    txtlines.append(delim.join(line))
                else:
                    pass
            else:
                txtlines += row.get_row_strings(column_widths, self.columns, delim, env)
        out = txtlines

        if self._title is not None:
            out = ['', self._title.get_str(env), ''] + out

        if self._caption:
            if self.rules:
                out.append(midrule)
            if isinstance(self._caption, str):
                out.append(self._caption)
            elif self._caption:
                out.append(str(self._caption))

        return linesep.join(out)

    def get_tex(self, env=None):
        if env is None:
            return tex(self)
        standalone = env.get('standalone')  # https://stackoverflow.com/a/17235546
        # init
        items = []
        if not standalone:
            items.append(r"\begin{center}")
        items.append(r"\begin{tabular}{%s}" % self.columns)
        if self.rules:
            items.append(r"\toprule")
        # Body
        for row in self.rows:
            if isinstance(row, str):
                items.append(row)
            else:
                items.append(row.get_tex(env))
        # post
        if self.rules:
            items.append(r"\bottomrule")
        items.append(r"\end{tabular}")
        if not standalone:
            items.append(r"\end{center}")
        return '\n'.join(items)

    def get_tsv(self, delimiter='\t', fmt='%.9g') -> str:
        r"""Render the table to tab-separated values (TSV) string

        Parameters
        ----------
        delimiter : str
            Delimiter between columns (default ``'\t'``).
        fmt : str
            Format string for numerical entries (default ``'%.9g'``).
        """
        buffer = StringIO(newline='')
        writer = csv.writer(buffer, delimiter=delimiter, lineterminator='\n')
        env = {'fmt': fmt}
        for row in self.rows:
            if isinstance(row, str):
                pass
            else:
                writer.writerow([cell.get_str(env) for cell in row])
        return buffer.getvalue()

    def save_docx(self, path: PathArg = None):
        """Save table as *.docx (requires `python-docx <https://python-docx.readthedocs.io>`_)

        Parameters
        ----------
        path : str
            Target file name (leave unspecified to use Save As dialog).

        Notes
        -----
        Most style options are not implemented.
        """
        try:
            from docx import Document
        except ImportError:
            raise ImportError("The python-docx package needs to be installed for this function")

        if path is None:
            path = ui.ask_saveas("Save Text File", filetypes=[("Word Document (*.docx)", "*.docx")])
        if not path:
            return
        document = Document()
        if self._title:
            document.add_heading(str(self._title), 0)

        table = document.add_table(rows=0, cols=self.n_columns)

        # Body
        for row in self.rows:
            if isinstance(row, str):
                continue
            doc_row = table.add_row()
            i = 0
            for cell in row:
                doc_cell = doc_row.cells[i]
                for j in range(i + 1, i + cell.width):
                    doc_cell.merge(doc_row.cells[j])
                doc_cell.text = str(cell)
                i += cell.width

        document.save(path)

    def save_tsv(self, path: PathArg = None, delimiter='\t', fmt='%.15g'):
        r"""Save the table as tab-separated values file

        Parameters
        ----------
        path
            Destination file name.
        delimiter : str
            String that is placed between cells (default: ``'\t'``).
        fmt : str
            Format string for representing numerical cells.
            (see 'Python String Formatting Documentation
            <http://docs.python.org/library/stdtypes.html#string-formatting-operations>'_)
        """
        _save_txt(self.get_tsv(delimiter, fmt), path)

    def save_txt(self, path: PathArg = None, fmt='%.15g', delim='   ', linesep='\n'):
        r"""Save the table as text file

        Parameters
        ----------
        path
            Destination file name.
        fmt : str
            Format string for representing numerical cells (default '%.15g').
        delim : str
            Cell delimiter.
        linesep : str
            String that is placed in between lines (default is ``'\n'``).
        """
        _save_txt(self.get_str({'fmt': fmt}, delim, linesep), path)


class Image(FMTextElement, BytesIO):
    "Represent an image file"

    def __init__(self, name=None, format='png', alt=None, buf=b'', height=None, width=None):
        """Represent an image file

        Parameters
        ----------
        name : str
            Name for the file (without extension; default is 'image').
        format : str
            File format (default 'png').
        alt : None | str
            Alternate text, placeholder in case the image can not be found
            (HTML `alt` tag, default is ``name``).
        buf : bytes
            Image buffer (optional).
        height : int
            Target height of the image; currently only used for iPython display.
        width : int
            Target width of the image; currently only used for iPython display.
        """
        BytesIO.__init__(self, buf)
        self.name = name or 'image'
        self.format = format
        self._alt = alt or self.name
        self._filename = os.extsep.join((self.name, format))
        self.height = height
        self.width = width

    @classmethod
    def from_array(cls, array, name='array', format='png', alt=None, height=None, width=None):
        """Create an Image object from an array.

        Parameters
        ----------
        array : array_like
            RGBA image array.
        name : None | str
            Name for the target image.
        format : str
            Format to save (default ``'png'``).
        alt : None | str
            Alternate text, placeholder in case the image can not be found
            (HTML `alt` tag).
        height : int
            Target height of the image; currently only used for iPython display.
        width : int
            Target width of the image; currently only used for iPython display.
        """
        im = cls(name, format, alt, height=height, width=width)
        imsave(im, np.asarray(array), format=format)
        return im

    @classmethod
    def from_file(cls, path: PathArg, name: str = None, alt: str = None):
        """Create an Image object from an existing image file.

        Parameters
        ----------
        path
            Path to the image file.
        name
            Name for the file (the default is ``os.path.basename(path)`` without
            extension.
        alt
            Alternate text, placeholder in case the image can not be found
            (HTML `alt` tag).
        """
        if name is None:
            filename = os.path.basename(path)
            name, ext = os.path.splitext(filename)
        else:
            _, ext = os.path.splitext(path)

        with open(path, 'rb') as fid:
            buf = fid.read()

        return cls(name, ext[1:], alt, buf)

    def _repr_items(self):
        out = [repr(self._filename)]
        if self._alt != self._filename:
            out.append(repr(self._alt))
        v = self.getvalue()
        if len(v) > 0:
            out.append('buf=%s...' % repr(v[:50]))
        return out

    def _repr_png_(self):  # iPython display method
        if self.height and self.width:
            meta = {'height': self.height, 'width': self.width}
        else:
            meta = {}
        return base64.b64encode(self.getvalue()).decode(), meta

    def get_html(self, env: dict = ENV):
        resource_dir = env.get('resource_dir', None)
        if resource_dir is None:
            buf = self.getvalue()
            if self.format == 'svg':
                return buf.decode()
            # http://stackoverflow.com/a/7389616/166700
            data = base64.b64encode(buf).decode().replace('\n', '')
            src = f'data:image/{self.format};base64,{data}'
        else:
            root = env['root']
            dst = root / resource_dir / self._filename
            if dst.exists():
                i = 0
                name, ext = os.path.splitext(self._filename)
                while dst.exists():
                    i += 1
                    dst = root / resource_dir / f'{name}-{i}{ext}'
            self.save_image(dst)
            src = dst.relative_to(root)
        alt = html(self._alt)
        return f'<img src="{src}" alt="{alt}">'

    def get_str(self, env: dict = ENV):
        return f"Image ({self._alt})"

    def save_image(self, dst):
        if os.path.isdir(dst):
            dst = os.path.join(dst, self._filename)

        buf = self.getvalue()
        with open(dst, 'wb') as fid:
            fid.write(buf)


class Figure(FMText):
    "Represent a figure with figure caption"

    def __init__(
            self,
            content: FMTextLike,
            caption: FMTextLike = None,
            options: dict = None,
            rasterize: bool = None,
    ):
        """Represent a figure

        Parameters
        ----------
        content
            Figure content.
        caption
            Figure caption.
        options
            HTML options for ``<figure>`` tag.
        rasterize
            Prefer rasterized graphics over vector graphics (applies when
            processing ``content``).
        """
        self._caption = asfmtext_or_none(caption, rasterize=rasterize)
        FMText.__init__(self, content, None, options, rasterize)

    def get_html(self, env: dict = ENV):
        body = FMText.get_html(self, env)
        if self._caption:
            caption = _html_element('figcaption', self._caption, env)
            body = '\n'.join((body, caption))
        return _html_element('figure', body, env, self.options)

    def get_str(self, env: dict = ENV):
        body = FMText.get_str(self, env)
        if self._caption:
            caption = str(self._caption)
            return "%s\n\n%s\n" % (body, caption)
        else:
            return body + '\n'


class FloatingLayout(FMText):
    """Arrange contents as floating elements

    Parameters
    ----------
    content
        Floating layout contents.
    rasterize
        Try to rasterize images in ``content``.
    """

    def __init__(self, content: FMTextLike, rasterize: bool = False):
        FMText.__init__(self, content, rasterize=rasterize)
        self._options = {
            'display': 'inline-block',
            'margin': '10px',
        }

    def get_html(self, env):
        options = (f"{key}: {value};" for key, value in self._options.items())
        header = '\n'.join((
            "<style> .floating-box {",
            *options,
            "} </style>",
        ))
        items = (item.get_html(env) for item in self.content)
        items = (f'<div class="floating-box">\n{item}\n</div>' for item in items)
        return '\n\n'.join([header, *items])


class Section(FMText):
    """Document section containing a title and content

    Parameters
    ----------
    heading : FMTextLike
        Section heading.
    content : FMTextLike
        Section content. Can also be constructed dynamically through the
        different .add_... methods.
    """
    def __init__(
            self,
            heading: FMTextLike,
            content: FMTextLike = None,
    ):
        self._heading = asfmtext(heading)
        FMText.__init__(self, content)

    def add_figure(
            self,
            caption: FMTextLike,
            content: FMTextLike = None,
            options: dict = None,
    ):
        """Add a figure frame to the section

        Parameters
        ----------
        caption : FMTextLike
            Figure caption.
        content : FMTextLike
            Figure content.
        options : dict
            HTML options for ``<figure>`` tag.

        Returns
        -------
        figure : Figure
            Figure object that was added.

        See Also
        --------
        .add_image_figure: add figure with a single image
        """
        if content is None:
            content = []
        figure = Figure(content, caption, options)
        self.append(figure)
        return figure

    def add_image_figure(
            self,
            image: Union[Image, np.ndarray, FMTextArg],
            caption: FMTextArg,
            alt: str = None,
    ):
        """Add an image in a figure frame to the section

        Parameters
        ----------
        image
            Image, image array or target filename for the image. If a filename
            it should have the appropriate extension.
        caption
            Figure caption.
        alt
            Alternate text, placeholder in case the image can not be found
            (HTML `alt` tag).

        Returns
        -------
        image : Image
            Image object that was added.

        See Also
        --------
        .add_figure: add an empty figure

        Notes
        -----
        Return the Image object which can be used to write the image data. If
        a function supports writing to a file-like object, it can be written
        with ``save_image(image)``. A function that writes a file to disk can
        be used as ``save_image(image.``.

        """
        if isinstance(image, str):
            name, ext = os.path.splitext(image)
            if ext:
                format = ext[1:]
            else:
                format = 'png'
            image = Image(name, format, alt)
        elif isinstance(image, np.ndarray):
            image = Image.from_array(image, alt=alt)
        else:
            image = asfmtext(image)

        figure = Figure(image, caption)
        self.append(figure)
        return image

    def add_paragraph(self, content=[]):
        paragraph = FMText(content, 'paragraph')
        self.append(paragraph)
        return paragraph

    def add_section(
            self,
            heading: FMTextLike,
            content: FMTextLike = ()):
        """Add a new subordinate section

        Parameters
        ----------
        heading : FMTextLike
            Heading for the section.
        content : FMTextLike
            Content for the section.

        Returns
        -------
        section : Section
            The new section.
        """
        section = Section(heading, content)
        self.append(section)
        return section

    def get_html(self, env: dict = ENV):
        env = env.copy()

        heading = self._get_html_section_heading(env)
        body = FMText.get_html(self, env)
        return '\n\n'.join(('', heading, body))

    def _get_html_section_heading(self, env):
        heading = self._heading.get_html(env)

        level = env.get('level', 1)
        tag = 'h%i' % level
        txt = _html_element(tag, heading, env)
        if 'toc' in env:
            toc_id = max(env['toc_ids']) + 1
            env['toc_ids'].append(toc_id)
            txt = _html_element('a', txt, env, {'name': toc_id})

            toc_txt = _html_element('a', heading, env, {'href': '#%i' % toc_id})
            env['toc'].append((level, toc_txt))
        env['level'] = level + 1
        return txt

    def get_str(self, env: dict = ENV):
        level = env.get('level', (1,))
        number = '.'.join(map(str, level))
        title = ' '.join((number, str(self._heading)))
        if len(level) == 1:
            underline_char = '='
        else:
            underline_char = '-'
        underline = underline_char * len(title)

        content = [title, underline, '']
        env = env.copy()
        level = list(level) + [1]
        for item in self.content:
            if isinstance(item, Section):
                env['level'] = tuple(level)
                txt = item.get_str(env)
                level[-1] += 1
                content += ['', '', txt]
            else:
                content += [str(item)]

        txt = '\n'.join(content)
        return txt


class Report(Section):
    """Document consisting of several sections plus a title

    Parameters
    ----------
    title : FMTextLike
        Document title.
    author : FMTextLike
        Document autho.
    date : bool | FMTextLike
        Date to print on the report. If True (default), the current day
        (object initialization) is used.
    content : FMTextLike
        Report content. Can also be constructed dynamically through the
        different .add_... methods.
    site_title : str
        Set the HTML site title (the default is the same as title).
    """
    def __init__(
            self,
            title: FMTextLike,
            author: FMTextLike = None,
            date: Union[FMTextLike, bool] = True,
            content: FMTextLike = None,
            site_title: str = None,
    ):
        if author is not None:
            author = asfmtext(author, r'\author')
        if date:
            if date is True:
                date = str(datetime.date.today())
            elif isinstance(date, str) and '%' in date:
                date = datetime.datetime.now().strftime(date)
            date = asfmtext(date, r'\date')
        self._author = author
        self._date = date
        self._site_title = site_title
        Section.__init__(self, title, content)

    def _repr_items(self):
        out = list(map(repr, (self._heading, self._author, self._date, self.content)))
        if self._site_title:
            out.append(repr(self._site_title))
        return out

    def get_html(self, env: dict = ENV):
        # setup TOC in env
        env = env.copy()
        env['toc'] = []
        env['toc_ids'] = [-1]
        env['level'] = 2

        # format document body (& collect document info)
        body = FMText.get_html(self, env)

        # format TOC
        toc = ['<ul>']
        level = 2
        for item_level, item in env['toc']:
            if item_level > level:
                toc.append('<ul>' * (item_level - level))
            elif item_level < level:
                toc.append('</ul>' * (level - item_level))
            toc.append(_html_element('li', item, env))
            level = item_level
        toc.append('</ul></li>' * (level - 1) + '</ul>')
        toc = '\n'.join(toc)

        # compile document content
        content = []
        title = _html_element('h1', self._heading, env)
        content.append(title)
        if self._author is not None:
            author = html(self._author, env)
            content.append(author)
        if self._date:
            date = html(self._date, env)
            content.append(date)
        content.append(toc)
        content.append(body)
        txt = '\n<br>\n'.join(content)
        return txt

    def get_str(self, env: dict = ENV):
        content = []
        title = str(self._heading)
        underline = '^' * len(title)
        content += [title, underline, '']
        if self._author is not None:
            author = self._author.get_str(env)
            content += [author, '']
        if self._date:
            date = self._date.get_str(env)
            content += [date, '']

        level = [1]
        env = env.copy()
        for item in self.content:
            if isinstance(item, Section):
                env['level'] = tuple(level)
                txt = item.get_str(env)
                level[-1] += 1
            else:
                txt = item.get_str(env)
            content += [txt, '']

        txt = '\n'.join(content)
        return txt

    def save_html(self, path, embed_images=True, meta=None):
        """Save HTML file of the report

        Parameters
        ----------
        path : str
            Path at which to save the html. Does not need to contain the
            extension. A folder with the same name is created fo resource
            files.
        embed_images : bool
            Embed images in the HTML file (default True). If False, a separate
            folder containing image files is created.
        meta : dict
            Meta-information for document head.
        """
        save_html(self, path, embed_images, meta)

    def show(self):
        "Save the report as temporary file and open in browser"
        temp_dir = tempfile.mkdtemp()
        path = os.path.join(temp_dir, 'report.html')
        self.save_html(path)
        webbrowser.open('file:/' + os.path.realpath(path))

    def sign(self, packages=('eelbrain',)):
        """Add a signature to the report

        Parameters
        ----------
        packages : iterator of str
            Packages whose ``__version__`` to include in the signature (default
            is eelbrain only).

        Notes
        -----
        Includes informationon on the computer that created the report, time
        of creation, and version of selected packages.
        """
        info = ["Created by %s on %s" % (socket.gethostname(),
                                         time.strftime("%c"))]
        for package in packages:
            if package in sys.modules:
                mod = sys.modules[package]
            else:
                continue

            if package == 'eelbrain':
                name = 'Eelbrain'
            elif package == 'mne':
                name = 'MNE-Python'
            elif package == 'surfer':
                name = 'PySurfer'
            else:
                name = package
            info.append(f"{name} version {mod.__version__}")

        signature = ' \u2014 \n'.join(info)
        self.add_paragraph(signature)


def symbol(symbol, subscript=None, tag='math'):
    if subscript is None:
        return Text(symbol, tag)
    else:
        return FMText([Text(symbol), Text(subscript, '_')], tag)


def p(p, stars=None, of=3, tag='math'):
    """:class:`FMText` representation of a p-value

    Parameters
    ----------
    p : scalar
        P-value.
    stars : str | True
        Stars decorating the p-value (e.g., "**"); ``True`` to use
        conventional thresholds (default is no stars).
    of : int
        Max numbers of star characters possible (to add empty space for
        alignment; ignored when ``stars`` is not specified).

    Returns
    -------
    text : FMText
        FMText with formatted p-value.
    """
    if stars is None:
        return P(p)
    elif stars is True:
        stars_obj = Stars.from_p(p, of)
    else:
        stars_obj = Stars(stars, of)
    return FMText([P(p), stars_obj], tag)


def stat(x, fmt="%.2f", stars=None, of=3, tag='math', drop0=False):
    ":class:`FMText` with properties for a statistic (e.g. a t-value)"
    if stars is None:
        return Number(x, tag, fmt, drop0)
    elif isinstance(stars, float):
        stars_obj = Stars.from_p(stars, of)
    else:
        stars_obj = Stars(stars, of)
    return FMText([Number(x, None, fmt, drop0), stars_obj], tag)


def eq(name, result, subscript=None, fmt='%.2f', stars=None, of=3, drop0=False):
    "``$name_{subscript} = result^stars$``"
    symbol_ = symbol(name, subscript, None)
    stat_ = stat(result, fmt, stars, of, None, drop0)
    return FMText([symbol_, Text(' = '), stat_], 'math')


def peq(p, subscript=None, stars=None, of=3):
    """:class:`FMText` representation of a p-value equation

    Parameters
    ----------
    p : scalar
        P-value.
    subscript : Text
        Subscript for ``p``.
    stars : str | True
        Stars decorating the p-value (e.g., "**"); ``True`` to use
        conventional thresholds (default is no stars).
    of : int
        Max numbers of star characters possible (to add empty space for
        alignment; ignored when ``stars`` is not specified).

    Returns
    -------
    text : FMText
        FMText with ``p = ...`` equation.
    """
    symbol_ = symbol('p', subscript, None)

    if p < .001:
        eq_ = Text(' ')
    else:
        eq_ = Text(' = ')

    if stars is None:
        return FMText([symbol_, eq_, P(p)], 'math')
    elif stars is True:
        stars_obj = Stars.from_p(p, of)
    else:
        stars_obj = Stars(stars, of)
    return FMText([symbol_, eq_, P(p), stars_obj], 'math')


def delim_list(items, delimiter=', '):
    delim = asfmtext(delimiter)
    out = list(items)
    for i in range(len(out) - 1, 0, -1):
        out.insert(i, delim)
    return out


def unindent(text, skip1=False):
    """Remove leading spaces that are present in all lines of ``text``.

    Parameters
    ----------
    test : str
        The text from which leading spaces should be removed.
    skip1 : bool
        Ignore the first line when determining number of spaces to unindent,
        and remove all leading whitespaces from it.
    """
    # count leading whitespaces
    lines = text.splitlines()
    ws_lead = []
    for line in lines[skip1:]:
        len_stripped = len(line.lstrip(' '))
        if len_stripped:
            ws_lead.append(len(line) - len_stripped)

    if len(ws_lead) > skip1:
        rm = min(ws_lead)
        if rm:
            if skip1:
                lines[0] = ' ' * rm + lines[0].lstrip()

            text = '\n'.join(line[rm:] for line in lines)

    return text


def im_table(ims, header=None, name="im_table"):
    """Create an SVG with a table of images

    Parameters
    ----------
    ims : list of list of arrays
        Images. Should all have same shape.
    header : None | list of str
        List of column titles.

    Returns
    -------
    images : Image
        FMTXT Image object that can be saved as SVG or integrated into an
        FMTXT document.
    """
    lens = set(map(len, ims))
    if len(lens) > 1:
        raise ValueError("Unequal number of columns")
    n_cols = lens.pop()
    n_rows = len(ims)

    # test im shape
    im0 = ims[0][0]
    if im0.ndim != 3:
        raise ValueError("Images need ndim=3, got %i" % ims[0][0].ndim)
    shape = im0.shape
    if not all(im.shape == shape for line in ims for im in line):
        raise NotImplementedError("Not all images have same shape")
    im_h, im_w, _ = shape

    xs = list(range(0, im_w * n_cols, im_w))
    y0 = 0 if header is None else 25
    ys = list(range(y0, y0 + im_h * n_rows, im_h))

    svg_h = y0 + im_h * n_rows
    svg_w = im_w * n_cols
    svg = ['<svg width="{w}" height="{h}">'.format(w=svg_w, h=svg_h)]
    if header is not None:
        assert len(header) == n_cols
        p = '<text x="{x}" y="{y}">{txt}</text>'
        for x, txt in zip(xs, header):
            item = p.format(x=x, y=18, txt=txt)
            svg.append(item)

    for y, line in zip(ys, ims):
        for x, im in zip(xs, line):
            data = _array_as_png(im)
            item = f'<image x="{x}" y="{y}" width="{im_w}" height="{im_h}" xlink:href="data:image/png;base64,{data}" />'
            svg.append(item)

    svg.append("</svg>")
    buf = '\n'.join(svg)
    return Image(name, 'svg', buf=buf.encode())


def _array_as_png(im):
    """Convert array to base64 encoded PNG

    Parameters
    ----------
    im : array_like
        An MxN (luminance), MxNx3 (RGB) or MxNx4 (RGBA) array.

    Returns
    -------
    data : str
        Encoded PNG image.
    """
    bytesio = BytesIO()
    imsave(bytesio, np.asarray(im), format='png')
    buf = bytesio.getvalue()
    data = base64.b64encode(buf).decode().replace('\n', '')
    return data


class MetaParser(HTMLParser):
    def __init__(self, file_name):
        HTMLParser.__init__(self)
        self.out = None
        with open(file_name) as fid:
            while self.out is None:
                self.feed(fid.readline())

    def handle_starttag(self, tag, attrs):
        if tag == 'meta':
            self.out = dict(attrs)

    def handle_endtag(self, tag):
        "No need to keep reading"
        if tag == 'head':
            self.out = {}


def read_meta(file_path):
    "Read meta information from a HTML head"
    return MetaParser(file_path).out


FMTextArg = Union[str, ListType[str], FMTextElement, FMTextConstant]



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from itertools import repeat
from math import floor
from typing import Literal

from eelbrain import NDVar, Scalar, UTS
from eelbrain._utils import tqdm
from numba import njit, prange
import numpy


@njit(parallel=True)
def aggregate_left(xf: numpy.ndarray, n_samples: int, step: float, n_window: int):
    xf **= 2
    out = numpy.empty(n_samples)
    for i in prange(n_samples):
        start = int(round(i * step))
        out[i] = xf[start: start + n_window].mean()
    return out


@njit(parallel=True)
def aggregate_right(xf: numpy.ndarray, n_samples: int, step: float, n_window: int):
    xf **= 2
    out = numpy.empty(n_samples)
    for i in prange(n_samples):
        stop = int(round(i * step)) + 1
        start = stop - n_window
        if start < 0:
            start = 0
        out[i] = xf[start: stop].mean()
    return out


def gammatone_bank(
        wav: NDVar,
        f_min: float,
        f_max: float,
        n: int,
        tstep: float = None,
        integration_cycles: float = None,
        integration_window: float = None,
        location: Literal['left', 'right', 'center'] = 'right',
        name: str = None,
) -> NDVar:
    """Gammatone filterbank response

    Parameters
    ----------
    wav
        Sound input.
    f_min
        Lower frequency cutoff.
    f_max
        Upper frequency cutoff.
    n
        Number of filter channels.
    tstep
        Time step size in the output (default is same as ``wav``).
    integration_cycles
        Number of cycles over which to integrate filter output (default 2).
    integration_window
        Integration time window in seconds (can be specified as alternative to
        ``integration_cycles``; e.g. ``0.010`` for 10 ms).
    location
        Location of the output relative to the input time axis:

        - ``right``: gammatone sample at end of integration window (default)
        - ``left``: gammatone sample at beginning of integration window
        - ``center``: gammatone sample at center of integration window

        Since gammatone filter response depends on ``integration_window``, the
        filter response will be delayed relative to the analytic envlope. To
        prevent this delay, use `location='left'`
    pad
        Pad output to match time axis of input.
    name
        NDVar name (default is ``wav.name``).

    Notes
    -----
    This function uses the `<https://github.com/detly/gammatone>` library, which
    is a Python port of Malcolm Slaney's and Dan Ellis' MATLAB gammatone
    filterbank.
    It requires the ``fmax`` bug-fix branch, which can be instaled with:

        $ pip install https://github.com/christianbrodbeck/gammatone/archive/fmax.zip
    """
    try:
        from gammatone.filters import centre_freqs, erb_filterbank
        from gammatone.gtgram import make_erb_filters
    except ModuleNotFoundError:
        raise ModuleNotFoundError("gammatone module not installed. Install using:\n\n  $ pip install https://github.com/christianbrodbeck/gammatone/archive/fmax.zip")

    assert location in ('left', 'right')
    fs = 1 / wav.time.tstep
    cfs = centre_freqs(fs, n, f_min, f_max)
    if integration_window is None:
        if integration_cycles is None:
            integration_cycles = 2
        integration_window_sec = integration_cycles / cfs
        integration_window_lens = numpy.ceil(integration_window_sec / wav.time.tstep).astype(int)
    elif integration_cycles is not None:
        raise ValueError(f"{integration_cycles=}, {integration_window=} (parameters are mutually exclusive")
    else:
        integration_window_len = int(round(integration_window * fs))
        integration_window_lens = repeat(integration_window_len)

    wav_ = wav
    tmin = wav.time.tmin
    if tstep is None:
        tstep = wav.time.tstep
    wave = wav_.get_data('time')
    # based on gammatone library gtgram, rewritten to reduce memory footprint
    output_n_samples = floor(len(wave) * wav.time.tstep / tstep)
    output_step = tstep / wav.time.tstep
    output_data = numpy.empty((n, output_n_samples))
    disable = n * output_n_samples < 200_000  # 100 bands * 2 s * 1000 Hz
    for i, cf, window in tqdm(zip(range(n-1, -1, -1), cfs, integration_window_lens), "Gammatone filterbank", total=len(cfs), unit='band', disable=disable):
        fcoefs = numpy.flipud(make_erb_filters(fs, cf))
        xf = erb_filterbank(wave, fcoefs)
        if location == 'left':
            output_data[i] = aggregate_left(xf[0], output_n_samples, output_step, window)
        else:
            output_data[i] = aggregate_right(xf[0], output_n_samples, output_step, window)
    output_data = numpy.sqrt(output_data, out=output_data)
    # package output
    freq_dim = Scalar('frequency', cfs[::-1], 'Hz')
    time_dim = UTS(tmin, tstep, output_n_samples)
    if name is None:
        name = wav.name
    return NDVar(output_data, (freq_dim, time_dim), name)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from eelbrain.pipeline import PrimaryEpoch, SecondaryEpoch, SuperEpoch, EpochCollection, ContinuousEpoch


def test_epoch_repr():
    primary_epoch = PrimaryEpoch('session')
    assert repr(primary_epoch) == "PrimaryEpoch('session', samplingrate=200, baseline=(None, 0))"
    secondary_epoch = SecondaryEpoch('primary_epoch', 'v == 1')
    assert repr(secondary_epoch) == "SecondaryEpoch('primary_epoch', 'v == 1')"
    super_epoch = SuperEpoch(('e1', 'e2'))
    assert repr(super_epoch) == "SuperEpoch(('e1', 'e2'))"
    epoch_collection = EpochCollection(('e1', 'e2'))
    assert repr(epoch_collection) == "EpochCollection(('e1', 'e2'))"
    continuous_epoch = ContinuousEpoch('session', 'stim == 1')
    assert repr(continuous_epoch) == "ContinuousEpoch('session', sel='stim == 1')"
